import matplotlib.pyplot as plt
import matplotlib.ticker as mticker
import numpy as np
import pandas as pd
from statsmodels.compat.python import zip_longest
from statsmodels.iolib.table import SimpleTable
from statsmodels.iolib.tableformatting import fmt_2cols

import acrl.synthetic as sca


def generate_table(left_col, right_col, table_title):
    col_headers = None

    if right_col:
        if len(right_col) < len(left_col):
            right_col += [(" ", " ")] * (len(left_col) - len(right_col))
        elif len(right_col) > len(left_col):
            left_col += [(" ", " ")] * (len(right_col) - len(left_col))

        right_col = [("%-21s" % ("  " + k), v) for k, v in right_col]

        gen_stubs_right, gen_data_right = zip_longest(*right_col)
        gen_table_right = SimpleTable(
            gen_data_right,
            col_headers,
            gen_stubs_right,
            title=table_title,
            txt_fmt=fmt_2cols,
        )

    else:
        gen_table_right = []

    gen_stubs_left, gen_data_left = zip_longest(*left_col)
    gen_table_left = SimpleTable(
        gen_data_left, col_headers, gen_stubs_left, title=table_title, txt_fmt=fmt_2cols
    )
    gen_table_left.extend_right(gen_table_right)
    general_table = gen_table_left

    return general_table


def get_env_param():
    env = sca.MarketEnvironment()

    fp_title = "Financial Parameters"
    fp_left_col = [
        ("Annual Volatility:", ["{:.0f}%".format(env.anv * 100)]),
        ("Daily Volatility:", ["{:.1f}%".format(env.dpv * 100)]),
    ]

    fp_right_col = [
        ("Bid-Ask Spread:", ["{:.3f}".format(env.basp)]),
        ("Daily Trading Volume:", ["{:,.0f}".format(env.dtv)]),
    ]

    acp_title = "Almgren and Chriss Model Parameters"
    acp_left_col = [
        ("Total Number of Shares to Sell:", ["{:,}".format(env.total_shares)]),
        ("Starting Price per Share:", ["${:.2f}".format(env.startingPrice)]),
        ("Price Impact for Each 1% of Daily Volume Traded:", ["${}".format(env.eta)]),
        ("Number of Days to Sell All the Shares:", ["{}".format(env.liquidation_time)]),
        ("Number of Trades:", ["{}".format(env.num_n)]),
    ]

    acp_right_col = [
        ("Fixed Cost of Selling per Share:", ["${:.3f}".format(env.epsilon)]),
        ("Trader's Risk Aversion:", ["{}".format(env.llambda)]),
        ("Permanent Impact Constant:", ["{}".format(env.gamma)]),
        ("Single Step Variance:", ["{:.3f}".format(env.singleStepVariance)]),
        ("Time Interval between trades:", ["{}".format(env.tau)]),
    ]

    fp_table = generate_table(fp_left_col, fp_right_col, fp_title)
    acp_table = generate_table(acp_left_col, acp_right_col, acp_title)

    return fp_table, acp_table


def plot_price_model(seed=0, num_days=1000):
    env = sca.MarketEnvironment()
    env.reset(seed)

    price_hist = np.zeros(num_days)

    for i in range(num_days):
        _, _, _, info = env.step(i)
        price_hist[i] = info.price

    print("Average Stock Price: ${:,.2f}".format(price_hist.mean()))
    print("Standard Deviation in Stock Price: ${:,.2f}".format(price_hist.std()))

    price_df = pd.DataFrame(data=price_hist, columns=["Stock"], dtype="float64")

    ax = price_df.plot(colormap="cool", grid=False)
    ax.set_facecolor(color="k")
    ax = plt.gca()

    yNumFmt = mticker.StrMethodFormatter("${x:,.2f}")

    ax.yaxis.set_major_formatter(yNumFmt)

    plt.ylabel("Stock Price")
    plt.xlabel("days")
    plt.show()


def get_optimal_vals(lq_time=60, nm_trades=60, tr_risk=1e-6, title=""):
    env = sca.MarketEnvironment()
    env.reset(liquid_time=lq_time, num_trades=nm_trades, lamb=tr_risk)

    if title == "":
        title = "AC Optimal Strategy"
    else:
        title = "AC Optimal Strategy for " + title

    E = env.get_AC_expected_shortfall(env.total_shares)
    V = env.get_AC_variance(env.total_shares)
    U = env.compute_AC_utility(env.total_shares)

    left_col = [
        ("Number of Days to Sell All the Shares:", ["{}".format(env.liquidation_time)]),
        ("Half-Life of The Trade:", ["{:,.1f}".format(1 / env.kappa)]),
        ("Utility:", ["${:,.2f}".format(U)]),
    ]

    right_col = [
        (
            "Initial Portfolio Value:",
            ["${:,.2f}".format(env.total_shares * env.startingPrice)],
        ),
        ("Expected Shortfall:", ["${:,.2f}".format(E)]),
        ("Standard Deviation of Shortfall:", ["${:,.2f}".format(np.sqrt(V))]),
    ]

    value_table = generate_table(left_col, right_col, title)

    return value_table


def get_min_param():
    min_impact = get_optimal_vals(
        lq_time=250, nm_trades=250, tr_risk=1e-17, title="Minimum Impact"
    )

    min_var = get_optimal_vals(
        lq_time=1, nm_trades=1, tr_risk=0.0058, title="Minimum Variance"
    )

    return min_impact, min_var


def get_crfs(trisk):
    tr_st = "{:.0e}".format(trisk)

    lnum = tr_st.split("e")[0]

    lexp = tr_st.split("e")[1]

    if np.abs(np.int(lexp)) < 10:
        lexp = lexp.replace("0", "", 1)

    an_st = "$\lambda = " + lnum + " \\times 10^{" + lexp + "}$"

    if trisk >= 1e-7 and trisk <= 4e-7:
        xcrf = 0.94
        ycrf = 2.5
        scrf = 0.1
    elif trisk > 4e-7 and trisk <= 9e-7:
        xcrf = 0.9
        ycrf = 2.5
        scrf = 0.06
    elif trisk > 9e-7 and trisk <= 1e-6:
        xcrf = 0.85
        ycrf = 2.5
        scrf = 0.06
    elif trisk > 1e-6 and trisk < 2e-6:
        xcrf = 1.2
        ycrf = 2.5
        scrf = 0.06
    elif trisk >= 2e-6 and trisk < 3e-6:
        xcrf = 0.8
        ycrf = 2.5
        scrf = 0.06
    elif trisk >= 3e-6 and trisk < 4e-6:
        xcrf = 0.7
        ycrf = 2.5
        scrf = 0.08
    elif trisk >= 4e-6 and trisk < 7e-6:
        xcrf = 1.4
        ycrf = 2.0
        scrf = 0.08
    elif trisk >= 7e-6 and trisk <= 1e-5:
        xcrf = 4.5
        ycrf = 1.5
        scrf = 0.08
    elif trisk > 1e-5 and trisk <= 2e-5:
        xcrf = 7.0
        ycrf = 1.1
        scrf = 0.08
    elif trisk > 2e-5 and trisk <= 5e-5:
        xcrf = 12.0
        ycrf = 1.1
        scrf = 0.08
    elif trisk > 5e-5 and trisk <= 1e-4:
        xcrf = 30
        ycrf = 0.99
        scrf = 0.08
    else:
        xcrf = 1
        ycrf = 1
        scrf = 0.08

    return an_st, xcrf, ycrf, scrf


def plot_efficient_frontier(tr_risk=1e-6):
    env = sca.MarketEnvironment()
    env.reset(lamb=tr_risk)

    tr_E = env.get_AC_expected_shortfall(env.total_shares)
    tr_V = env.get_AC_variance(env.total_shares)

    E = np.array([])
    V = np.array([])
    U = np.array([])

    num_points = 7000

    lambdas = np.linspace(1e-7, 1e-4, num_points)

    for llambda in lambdas:
        env.reset(lamb=llambda)
        E = np.append(E, env.get_AC_expected_shortfall(env.total_shares))
        V = np.append(V, env.get_AC_variance(env.total_shares))
        U = np.append(U, env.compute_AC_utility(env.total_shares))

    cm = plt.cm.get_cmap("gist_rainbow")

    sc = plt.scatter(V, E, s=20, c=U, cmap=cm)

    plt.colorbar(sc, label="AC Utility", format=mticker.StrMethodFormatter("${x:,.0f}"))

    ax = plt.gca()
    ax.set_facecolor("k")

    ymin = E.min() * 0.7
    ymax = E.max() * 1.1

    plt.ylim(ymin, ymax)

    yNumFmt = mticker.StrMethodFormatter("${x:,.0f}")
    xNumFmt = mticker.StrMethodFormatter("{x:,.0f}")

    ax.yaxis.set_major_formatter(yNumFmt)
    ax.xaxis.set_major_formatter(xNumFmt)

    plt.xlabel("Variance of Shortfall")
    plt.ylabel("Expected Shortfall")

    an_st, xcrf, ycrf, scrf = get_crfs(tr_risk)

    plt.annotate(
        an_st,
        xy=(tr_V, tr_E),
        xytext=(tr_V * xcrf, tr_E * ycrf),
        color="w",
        size="large",
        arrowprops=dict(facecolor="cyan", shrink=scrf, width=3, headwidth=10),
    )

    plt.show()


def round_trade_list(trl):
    trl_rd = np.around(trl)

    res = np.around(trl.sum() - trl_rd.sum())

    if res != 0:
        idx = trl_rd.nonzero()[0][-1]
        trl_rd[idx] += res

    return trl_rd


def plot_trade_list(lq_time=60, nm_trades=60, tr_risk=1e-6, show_trl=False):
    env = sca.MarketEnvironment()
    env.reset(liquid_time=lq_time, num_trades=nm_trades, lamb=tr_risk)

    trade_list = env.get_trade_list()

    new_trl = np.insert(trade_list, 0, 0)

    df = pd.DataFrame(
        data=list(range(nm_trades + 1)), columns=["Trade Number"], dtype="float64"
    )

    df["Stocks Sold"] = new_trl

    df["Stocks Remaining"] = (np.ones(nm_trades + 1) * env.total_shares) - np.cumsum(
        new_trl
    )

    _, axes = plt.subplots(nrows=1, ncols=2)

    df.iloc[1:].plot.scatter(
        x="Trade Number",
        y="Stocks Sold",
        c="Stocks Sold",
        colormap="gist_rainbow",
        alpha=1,
        sharex=False,
        s=50,
        colorbar=False,
        ax=axes[0],
    )

    axes[0].plot(
        df["Trade Number"].iloc[1:],
        df["Stocks Sold"].iloc[1:],
        linewidth=2.0,
        alpha=0.5,
    )

    axes[0].set_facecolor(color="k")

    yNumFmt = mticker.StrMethodFormatter("{x:,.0f}")

    axes[0].yaxis.set_major_formatter(yNumFmt)
    axes[0].set_title("Trading List")

    df.plot.scatter(
        x="Trade Number",
        y="Stocks Remaining",
        c="Stocks Remaining",
        colormap="gist_rainbow",
        alpha=1,
        sharex=False,
        s=50,
        colorbar=False,
        ax=axes[1],
    )

    axes[1].plot(df["Trade Number"], df["Stocks Remaining"], linewidth=2.0, alpha=0.5)
    axes[1].set_facecolor(color="k")

    yNumFmt = mticker.StrMethodFormatter("{x:,.0f}")

    axes[1].yaxis.set_major_formatter(yNumFmt)
    axes[1].set_title("Trading Trajectory")

    plt.subplots_adjust(wspace=0.4)
    plt.show()

    print("\nNumber of Shares Sold: {:,.0f}\n".format(new_trl.sum()))

    if show_trl:
        rd_trl = round_trade_list(new_trl)

        df2 = pd.DataFrame(
            data=list(range(nm_trades + 1)), columns=["Trade Number"], dtype="float64"
        )

        df2["Stocks Sold"] = rd_trl
        df2["Stocks Remaining"] = (
            np.ones(nm_trades + 1) * env.total_shares
        ) - np.cumsum(rd_trl)

        return df2.style.hide_index().format(
            {
                "Trade Number": "{:.0f}",
                "Stocks Sold": "{:,.0f}",
                "Stocks Remaining": "{:,.0f}",
            }
        )


def implement_trade_list(seed=0, lq_time=60, nm_trades=60, tr_risk=1e-6):
    env = sca.MarketEnvironment()
    env.reset(seed=seed, liquid_time=lq_time, num_trades=nm_trades, lamb=tr_risk)

    trl = env.get_trade_list()
    trade_list = round_trade_list(trl)

    env.start_transactions()

    price_hist = np.array([])

    for trade in trade_list:
        action = trade / env.shares_remaining

        _, _, _, info = env.step(action)

        price_hist = np.append(price_hist, info.exec_price)

        if info.done:
            print(
                "Implementation Shortfall: ${:,.2f} \n".format(
                    info.implementation_shortfall
                )
            )
            break

    price_df = pd.DataFrame(data=price_hist, columns=["Stock"], dtype="float64")

    ax = price_df.plot(colormap="cool", grid=False)
    ax.set_facecolor(color="k")
    ax.set_title("Impacted Stock Price")

    ax = plt.gca()

    yNumFmt = mticker.StrMethodFormatter("${x:,.2f}")

    ax.yaxis.set_major_formatter(yNumFmt)

    plt.plot(price_hist, "o")
    plt.ylabel("Stock Price")
    plt.xlabel("Trade Number")
    plt.show()


def get_av_std(lq_time=60, nm_trades=60, tr_risk=1e-6, trs=100):
    env = sca.MarketEnvironment()
    env.reset(liquid_time=lq_time, num_trades=nm_trades, lamb=tr_risk)

    trl = env.get_trade_list()
    trade_list = round_trade_list(trl)
    shortfall_hist = np.array([])

    for episode in range(trs):
        if (episode + 1) % 100 == 0:
            print("Episode [{}/{}]".format(episode + 1, trs), end="\r", flush=True)

        env.reset(seed=episode, liquid_time=lq_time, num_trades=nm_trades, lamb=tr_risk)
        env.start_transactions()

        for trade in trade_list:
            action = trade / env.shares_remaining
            _, _, _, info = env.step(action)

            if info.done:
                shortfall_hist = np.append(
                    shortfall_hist, info.implementation_shortfall
                )
                break

    print("Average Implementation Shortfall: ${:,.2f}".format(shortfall_hist.mean()))
    print(
        "Standard Deviation of the Implementation Shortfall: ${:,.2f}".format(
            shortfall_hist.std()
        )
    )

    plt.plot(shortfall_hist, "cyan", label="")
    plt.xlim(0, trs)

    ax = plt.gca()
    ax.set_facecolor("k")
    ax.set_xlabel("Episode", fontsize=15)
    ax.set_ylabel("Implementation Shortfall (US $)", fontsize=15)

    ax.axhline(shortfall_hist.mean(), 0, 1, color="m", label="Average")

    yNumFmt = mticker.StrMethodFormatter("${x:,.0f}")

    ax.yaxis.set_major_formatter(yNumFmt)

    plt.legend()
    plt.show



import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F


def hidden_init(layer):
    fan_in = layer.weight.data.size()[0]
    lim = 1.0 / np.sqrt(fan_in)

    return (-lim, lim)


class Actor(nn.Module):
    def __init__(self, state_size, action_size, seed, fc1_units=24, fc2_units=48):
        super(Actor, self).__init__()
        self.seed = torch.manual_seed(seed)
        self.fc1 = nn.Linear(state_size, fc1_units)
        self.fc2 = nn.Linear(fc1_units, fc2_units)
        self.fc3 = nn.Linear(fc2_units, action_size)
        self.reset_parameters()

    def reset_parameters(self):
        self.fc1.weight.data.uniform_(*hidden_init(self.fc1))
        self.fc2.weight.data.uniform_(*hidden_init(self.fc2))
        self.fc3.weight.data.uniform_(-3e-3, 3e-3)

    def forward(self, state):
        x = F.relu(self.fc1(state))
        x = F.relu(self.fc2(x))

        return F.tanh(self.fc3(x))


class Critic(nn.Module):
    def __init__(self, state_size, action_size, seed, fcs1_units=24, fc2_units=48):
        super(Critic, self).__init__()
        self.seed = torch.manual_seed(seed)
        self.fcs1 = nn.Linear(state_size, fcs1_units)
        self.fc2 = nn.Linear(fcs1_units + action_size, fc2_units)
        self.fc3 = nn.Linear(fc2_units, 1)
        self.reset_parameters()

    def reset_parameters(self):
        self.fcs1.weight.data.uniform_(*hidden_init(self.fcs1))
        self.fc2.weight.data.uniform_(*hidden_init(self.fc2))
        self.fc3.weight.data.uniform_(-3e-3, 3e-3)

    def forward(self, state, action):
        xs = F.relu(self.fcs1(state))

        x = torch.cat((xs, action), dim=1)
        x = F.relu(self.fc2(x))

        return self.fc3(x)



import copy
import random
from collections import deque, namedtuple

import numpy as np
import torch
import torch.nn.functional as F
import torch.optim as optim

from acrl.model import Actor, Critic

BUFFER_SIZE = int(1e4)
BATCH_SIZE = 128
GAMMA = 0.99
TAU = 1e-3
LR_ACTOR = 1e-4
LR_CRITIC = 1e-3
WEIGHT_DECAY = 0

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


class Agent:
    def __init__(self, state_size, action_size, random_seed):
        self.state_size = state_size
        self.action_size = action_size
        self.seed = random.seed(random_seed)

        self.actor_local = Actor(state_size, action_size, random_seed).to(device)
        self.actor_target = Actor(state_size, action_size, random_seed).to(device)
        self.actor_optimizer = optim.Adam(self.actor_local.parameters(), lr=LR_ACTOR)

        self.critic_local = Critic(state_size, action_size, random_seed).to(device)
        self.critic_target = Critic(state_size, action_size, random_seed).to(device)
        self.critic_optimizer = optim.Adam(
            self.critic_local.parameters(), lr=LR_CRITIC, weight_decay=WEIGHT_DECAY
        )

        self.noise = OuNoise(action_size, random_seed)
        self.memory = ReplayBuffer(action_size, BUFFER_SIZE, BATCH_SIZE, random_seed)

    def step(self, state, action, reward, next_state, done):
        self.memory.add(state, action, reward, next_state, done)

        if len(self.memory) > BATCH_SIZE:
            experiences = self.memory.sample()
            self.learn(experiences, GAMMA)

    def act(self, state, add_noise=True):
        state = torch.from_numpy(state).float().to(device)
        self.actor_local.eval()

        with torch.no_grad():
            action = self.actor_local(state).cpu().data.numpy()

        self.actor_local.train()

        if add_noise:
            action += self.noise.sample()

        action = (action + 1.0) / 2.0

        return np.clip(action, 0, 1)

    def reset(self):
        self.noise.reset()

    def learn(self, experiences, gamma):
        states, actions, rewards, next_states, dones = experiences

        actions_next = self.actor_target(next_states)
        Q_targets_next = self.critic_target(next_states, actions_next)
        Q_targets = rewards + (gamma * Q_targets_next * (1 - dones))
        Q_expected = self.critic_local(states, actions)
        critic_loss = F.mse_loss(Q_expected, Q_targets)

        self.critic_optimizer.zero_grad()
        critic_loss.backward()
        self.critic_optimizer.step()

        actions_pred = self.actor_local(states)
        actor_loss = -self.critic_local(states, actions_pred).mean()

        self.actor_optimizer.zero_grad()
        actor_loss.backward()
        self.actor_optimizer.step()

        self.soft_update(self.critic_local, self.critic_target, TAU)
        self.soft_update(self.actor_local, self.actor_target, TAU)

    def soft_update(self, local_model, target_model, tau):
        for target_param, local_param in zip(
            target_model.parameters(), local_model.parameters()
        ):
            target_param.data.copy_(
                tau * local_param.data + (1.0 - tau) * target_param.data
            )


class OuNoise:
    def __init__(self, size, seed, mu=0.0, theta=0.15, sigma=0.2):
        self.mu = mu * np.ones(size)
        self.theta = theta
        self.sigma = sigma
        self.seed = random.seed(seed)
        self.reset()

    def reset(self):
        self.state = copy.copy(self.mu)

    def sample(self):
        x = self.state

        dx = self.theta * (self.mu - x) + self.sigma * np.array(
            [random.random() for i in range(len(x))]
        )

        self.state = x + dx

        return self.state


class ReplayBuffer:
    def __init__(self, action_size, buffer_size, batch_size, seed):
        self.action_size = action_size
        self.memory = deque(maxlen=buffer_size)
        self.batch_size = batch_size

        self.experience = namedtuple(
            "Experience",
            field_names=["state", "action", "reward", "next_state", "done"],
        )

        self.seed = random.seed(seed)

    def add(self, state, action, reward, next_state, done):
        self.memory.append(self.experience(state, action, reward, next_state, done))

    def sample(self):
        experiences = random.sample(self.memory, k=self.batch_size)

        states = (
            torch.from_numpy(np.vstack([e.state for e in experiences if e is not None]))
            .float()
            .to(device)
        )

        actions = (
            torch.from_numpy(
                np.vstack([e.action for e in experiences if e is not None])
            )
            .float()
            .to(device)
        )

        rewards = (
            torch.from_numpy(
                np.vstack([e.reward for e in experiences if e is not None])
            )
            .float()
            .to(device)
        )

        next_states = (
            torch.from_numpy(
                np.vstack([e.next_state for e in experiences if e is not None])
            )
            .float()
            .to(device)
        )

        dones = (
            torch.from_numpy(
                np.vstack([e.done for e in experiences if e is not None]).astype(
                    np.uint8
                )
            )
            .float()
            .to(device)
        )

        return (states, actions, rewards, next_states, dones)

    def __len__(self):
        return len(self.memory)



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.ticker as mticker

import syntheticChrissAlmgren as sca

from statsmodels.iolib.table import SimpleTable
from statsmodels.compat.python import zip_longest
from statsmodels.iolib.tableformatting import fmt_2cols


def generate_table(left_col, right_col, table_title):
    
    # Do not use column headers
    col_headers = None
    
    # Generate the right table
    if right_col:
        # Add padding
        if len(right_col) < len(left_col):
            right_col += [(' ', ' ')] * (len(left_col) - len(right_col))
        elif len(right_col) > len(left_col):
            left_col += [(' ', ' ')] * (len(right_col) - len(left_col))
        right_col = [('%-21s' % ('  '+k), v) for k,v in right_col]
        
        # Generate the right table
        gen_stubs_right, gen_data_right = zip_longest(*right_col)
        gen_table_right = SimpleTable(gen_data_right,
                                          col_headers,
                                          gen_stubs_right,
                                          title = table_title,
                                          txt_fmt = fmt_2cols)
    else:
        # If there is no right table set the right table to empty
        gen_table_right = []

    # Generate the left table  
    gen_stubs_left, gen_data_left = zip_longest(*left_col) 
    gen_table_left = SimpleTable(gen_data_left,
                                 col_headers,
                                 gen_stubs_left,
                                 title = table_title,
                                 txt_fmt = fmt_2cols)

    
    # Merge the left and right tables to make a single table
    gen_table_left.extend_right(gen_table_right)
    general_table = gen_table_left

    return general_table


def get_env_param():
    
    # Create a simulation environment
    env = sca.MarketEnvironment()

    # Set the title for the financial parameters table
    fp_title = 'Financial Parameters'

    # Get the default financial parameters from the simulation environment
    fp_left_col = [('Annual Volatility:', ['{:.0f}%'.format(env.anv * 100)]),
                   ('Daily Volatility:', ['{:.1f}%'.format(env.dpv * 100)])]
    
    fp_right_col = [('Bid-Ask Spread:', ['{:.3f}'.format(env.basp)]),
                    ('Daily Trading Volume:', ['{:,.0f}'.format(env.dtv)])]

    # Set the title for the Almgren and Chriss Model parameters table
    acp_title = 'Almgren and Chriss Model Parameters'

    # Get the default Almgren and Chriss Model Parameters from the simulation environment
    acp_left_col = [('Total Number of Shares to Sell:', ['{:,}'.format(env.total_shares)]),
                    ('Starting Price per Share:', ['${:.2f}'.format(env.startingPrice)]),
                    ('Price Impact for Each 1% of Daily Volume Traded:', ['${}'.format(env.eta)]),                    
                    ('Number of Days to Sell All the Shares:', ['{}'.format(env.liquidation_time)]),
                    ('Number of Trades:', ['{}'.format(env.num_n)])]

    acp_right_col = [('Fixed Cost of Selling per Share:', ['${:.3f}'.format(env.epsilon)]),
                     ('Trader\'s Risk Aversion:', ['{}'.format(env.llambda)]),
                     ('Permanent Impact Constant:', ['{}'.format(env.gamma)]),
                     ('Single Step Variance:', ['{:.3f}'.format(env.singleStepVariance)]),
                     ('Time Interval between trades:', ['{}'.format(env.tau)])]

    # Generate tables with the default financial and AC Model parameters
    fp_table = generate_table(fp_left_col, fp_right_col, fp_title)
    acp_table = generate_table(acp_left_col, acp_right_col, acp_title)

    return fp_table, acp_table


def plot_price_model(seed = 0, num_days = 1000):
    
    # Create a simulation environment
    env = sca.MarketEnvironment()

    # Reset the enviroment with the given seed
    env.reset(seed)

    # Create an array to hold the daily stock price for the given number of days
    price_hist = np.zeros(num_days)

    # Get the simulated stock price movement from the environment
    for i in range(num_days):
        _, _, _, info = env.step(i)    
        price_hist[i] = info.price
    
    # Print Average and Standard Deviation in Stock Price
    print('Average Stock Price: ${:,.2f}'.format(price_hist.mean()))
    print('Standard Deviation in Stock Price: ${:,.2f}'.format(price_hist.std()))
#     print('Standard Deviation of Random Noise: {:,.5f}'.format(np.sqrt(env.singleStepVariance * env.tau)))
    
    # Plot the price history for the given number of days
    price_df = pd.DataFrame(data = price_hist,  columns = ['Stock'], dtype = 'float64')
    ax = price_df.plot(colormap = 'cool', grid = False)
    ax.set_facecolor(color = 'k')
    ax = plt.gca()
    yNumFmt = mticker.StrMethodFormatter('${x:,.2f}')
    ax.yaxis.set_major_formatter(yNumFmt)
    plt.ylabel('Stock Price')
    plt.xlabel('days')
    plt.show()
    

    
def get_optimal_vals(lq_time = 60, nm_trades = 60, tr_risk = 1e-6, title = ''):
    
    # Create a simulation environment
    env = sca.MarketEnvironment()

    # Reset the enviroment with the given parameters
    env.reset(liquid_time = lq_time, num_trades = nm_trades, lamb = tr_risk)

    # Set the title for the AC Optimal Strategy table
    if title == '':
        title = 'AC Optimal Strategy'
    else:
        title = 'AC Optimal Strategy for ' + title

    # Get the AC optimal values from the environment
    E = env.get_AC_expected_shortfall(env.total_shares)
    V = env.get_AC_variance(env.total_shares)
    U = env.compute_AC_utility(env.total_shares)

    left_col = [('Number of Days to Sell All the Shares:', ['{}'.format(env.liquidation_time)]),
                ('Half-Life of The Trade:', ['{:,.1f}'.format(1 / env.kappa)]),
                ('Utility:', ['${:,.2f}'.format(U)])]

    right_col = [('Initial Portfolio Value:', ['${:,.2f}'.format(env.total_shares * env.startingPrice)]),
                 ('Expected Shortfall:', ['${:,.2f}'.format(E)]),
                 ('Standard Deviation of Shortfall:', ['${:,.2f}'.format(np.sqrt(V))])]

    # Generate the table with the AC optimal values
    val_table = generate_table(left_col, right_col, title)

    return val_table


def get_min_param():
    
    # Get the minimum impact AC parameters
    min_impact = get_optimal_vals(lq_time = 250, nm_trades = 250, tr_risk = 1e-17, title = 'Minimum Impact')
    
    # Get the minimum variance AC parameters
    min_var = get_optimal_vals(lq_time = 1, nm_trades = 1, tr_risk = 0.0058, title = 'Minimum Variance')
    
    return min_impact, min_var
  
            
def get_crfs(trisk):
    
    # Create the annotation label
    tr_st = '{:.0e}'.format(trisk)   
    lnum = tr_st.split('e')[0]   
    lexp = tr_st.split('e')[1]
    if np.abs(np.int(lexp)) < 10:
        lexp = lexp.replace('0', '', 1)    
    an_st = '$\lambda = ' + lnum + ' \\times 10^{' + lexp + '}$'
    
    # Set the correction factors for the annotation label
    if trisk >= 1e-7 and trisk <= 4e-7:
        xcrf = 0.94
        ycrf = 2.5
        scrf = 0.1
    elif trisk > 4e-7 and trisk <= 9e-7:
        xcrf = 0.9
        ycrf = 2.5
        scrf = 0.06
    elif trisk > 9e-7 and trisk <= 1e-6:
        xcrf = 0.85
        ycrf = 2.5
        scrf = 0.06
    elif trisk > 1e-6 and trisk < 2e-6:
        xcrf = 1.2
        ycrf = 2.5
        scrf = 0.06
    elif trisk >= 2e-6 and trisk < 3e-6:
        xcrf = 0.8
        ycrf = 2.5
        scrf = 0.06
    elif trisk >= 3e-6 and trisk < 4e-6:
        xcrf = 0.7
        ycrf = 2.5
        scrf = 0.08
    elif trisk >= 4e-6 and trisk < 7e-6:
        xcrf = 1.4
        ycrf = 2.0
        scrf = 0.08
    elif trisk >= 7e-6 and trisk <= 1e-5:
        xcrf = 4.5
        ycrf = 1.5
        scrf = 0.08
    elif trisk > 1e-5 and trisk <= 2e-5:
        xcrf = 7.0
        ycrf = 1.1
        scrf = 0.08
    elif trisk > 2e-5 and trisk <= 5e-5:
        xcrf = 12.
        ycrf = 1.1
        scrf = 0.08
    elif trisk > 5e-5 and trisk <= 1e-4:
        xcrf = 30
        ycrf = 0.99
        scrf = 0.08
    else:
        xcrf = 1
        ycrf = 1
        scrf = 0.08
    
    return an_st, xcrf, ycrf, scrf
    

def plot_efficient_frontier(tr_risk = 1e-6):
    
    # Create a simulation environment
    env = sca.MarketEnvironment()
    
    # Reset the enviroment with the given trader's risk aversion
    env.reset(lamb = tr_risk)

    # Get the expected shortfall and corresponding variance for the given trader's risk aversion
    tr_E = env.get_AC_expected_shortfall(env.total_shares)
    tr_V = env.get_AC_variance(env.total_shares)
    
    # Create empty arrays to hold our values of E, V, and U
    E = np.array([])
    V = np.array([])
    U = np.array([])
    
    # Set the number of plot points for our frontier
    num_points = 7000
    
    # Set the values of the trader's risk aversion to plot
    lambdas = np.linspace(1e-7, 1e-4, num_points)
    
    # Calclate E, V, U for each value of llambda
    for llambda in lambdas:
        env.reset(lamb = llambda)
        E = np.append(E, env.get_AC_expected_shortfall(env.total_shares))
        V = np.append(V, env.get_AC_variance(env.total_shares))
        U = np.append(U, env.compute_AC_utility(env.total_shares))
        
    # Plot E vs V and use U for the colorbar    
    cm = plt.cm.get_cmap('gist_rainbow')    
    sc = plt.scatter(V, E, s = 20, c = U, cmap = cm)
    plt.colorbar(sc, label = 'AC Utility', format = mticker.StrMethodFormatter('${x:,.0f}'))
    ax = plt.gca()
    ax.set_facecolor('k')
    ymin = E.min() * 0.7
    ymax = E.max() * 1.1
    plt.ylim(ymin, ymax)
    yNumFmt = mticker.StrMethodFormatter('${x:,.0f}')
    xNumFmt = mticker.StrMethodFormatter('{x:,.0f}')
    ax.yaxis.set_major_formatter(yNumFmt)
    ax.xaxis.set_major_formatter(xNumFmt)
    plt.xlabel('Variance of Shortfall')
    plt.ylabel('Expected Shortfall')
    
    # Get the annotation label and the correction factors
    an_st, xcrf, ycrf, scrf = get_crfs(tr_risk)
    
    # Plot the annotation in the above plot
    plt.annotate(an_st, xy = (tr_V, tr_E), xytext = (tr_V * xcrf, tr_E  * ycrf), color = 'w', size = 'large', 
                 arrowprops = dict(facecolor = 'cyan', shrink = scrf, width = 3, headwidth = 10))
    plt.show()
    
    
def round_trade_list(trl):
    
    # Round the shares in the trading list
    trl_rd = np.around(trl)
        
    # Rounding the number of shares in the trading list sometimes results in selling more or less
    # shares than we have available. We calculate the difference between to total number of shares
    # sold in the original trading list and the number of shares sold in the rounded list.
    # This difference will be used to correct for rounding errors. 
    res = np.around(trl.sum() - trl_rd.sum())
        
    # Correct the number of shares sold due to rounding errors if necessary
    if res != 0:
        idx = trl_rd.nonzero()[0][-1]      
        trl_rd[idx] += res
        
    return trl_rd

    
def plot_trade_list(lq_time = 60, nm_trades = 60, tr_risk = 1e-6, show_trl = False):
    
    # Create simulation environment
    env = sca.MarketEnvironment()

    # Reset the environment with the given parameters
    env.reset(liquid_time = lq_time, num_trades = nm_trades, lamb = tr_risk)

    # Get the trading list from the environment
    trade_list = env.get_trade_list()
    
    # Add a zero at the beginning of the trade list to indicate that at time 0 we don't sell any stocks
    new_trl = np.insert(trade_list, 0, 0)

    # We create a dataframe with the trading list and trading trajectory
    df = pd.DataFrame(data = list(range(nm_trades + 1)),  columns = ['Trade Number'], dtype = 'float64')
    df['Stocks Sold'] = new_trl
    df['Stocks Remaining'] = (np.ones(nm_trades + 1) * env.total_shares) - np.cumsum(new_trl)

    # Create a figure with 2 plots in 1 row
    fig, axes = plt.subplots(nrows = 1, ncols = 2)
    
    # Make a scatter plot of the trade list
    df.iloc[1:].plot.scatter(x = 'Trade Number', y = 'Stocks Sold', c = 'Stocks Sold', colormap = 'gist_rainbow',
                                                 alpha = 1, sharex = False, s = 50, colorbar = False, ax = axes[0])
    
    # Plot a line through the points of the scatter plot of the trade list
    axes[0].plot(df['Trade Number'].iloc[1:], df['Stocks Sold'].iloc[1:], linewidth = 2.0, alpha = 0.5)
    axes[0].set_facecolor(color = 'k')
    yNumFmt = mticker.StrMethodFormatter('{x:,.0f}')
    axes[0].yaxis.set_major_formatter(yNumFmt)
    axes[0].set_title('Trading List')

    # Make a scatter plot of the number of stocks remaining after each trade
    df.plot.scatter(x = 'Trade Number', y = 'Stocks Remaining', c = 'Stocks Remaining', colormap = 'gist_rainbow',
                                                 alpha = 1, sharex = False, s = 50, colorbar = False, ax = axes[1])
    
    # Plot a line through the points of the scatter plot of the number of stocks remaining after each trade
    axes[1].plot(df['Trade Number'], df['Stocks Remaining'], linewidth = 2.0, alpha = 0.5)
    axes[1].set_facecolor(color = 'k')
    yNumFmt = mticker.StrMethodFormatter('{x:,.0f}')
    axes[1].yaxis.set_major_formatter(yNumFmt)
    axes[1].set_title('Trading Trajectory')
    
    # Set the spacing between plots
    plt.subplots_adjust(wspace = 0.4)
    plt.show()
    
    print('\nNumber of Shares Sold: {:,.0f}\n'.format(new_trl.sum()))
    
    if show_trl:
        
        # Since we are not selling fractional shares we round up the shares in the trading list
        rd_trl = round_trade_list(new_trl)
#         rd_trl = new_trl

        # We create a dataframe with the modified trading list and trading trajectory
        df2 = pd.DataFrame(data = list(range(nm_trades + 1)),  columns = ['Trade Number'], dtype = 'float64')
        df2['Stocks Sold'] = rd_trl
        df2['Stocks Remaining'] = (np.ones(nm_trades + 1) * env.total_shares) - np.cumsum(rd_trl)

        return df2.style.hide_index().format({'Trade Number': '{:.0f}', 'Stocks Sold': '{:,.0f}', 'Stocks Remaining': '{:,.0f}'})
#         return df2.style.hide_index().format({'Trade Number': '{:.0f}', 'Stocks Sold': '{:e}', 'Stocks Remaining': '{:e}'})
    

def implement_trade_list(seed = 0, lq_time = 60, nm_trades = 60, tr_risk = 1e-6):
    
    # Create simulation environment
    env = sca.MarketEnvironment()

    # Reset the environment with the given parameters
    env.reset(seed = seed, liquid_time = lq_time, num_trades = nm_trades, lamb = tr_risk)

    # Get the trading list from the environment
    trl = env.get_trade_list()
    
    # Since we are not selling fractional shares we round up the shares in the trading list
    trade_list = round_trade_list(trl)
 
    # set the environment to make transactions
    env.start_transactions()
    
    # Create an array to hold the impacted stock price
    price_hist = np.array([])

    # Implement the trading list in our similation environment
    for trade in trade_list:
        
        # Convert the number of shares to sell in each trade into an action
        action = trade / env.shares_remaining
        
        # Take a step in the environment my selling the number of shares in the current trade
        _, _, _, info = env.step(action)
        
        # Get the impacted price from the environment
        price_hist = np.append(price_hist, info.exec_price)
        
        # If all shares have been sold, stop making transactions and get the implementation sortfall
        if info.done:
            print('Implementation Shortfall: ${:,.2f} \n'.format(info.implementation_shortfall))
            break

    # Plot the impacted price
    price_df = pd.DataFrame(data = price_hist,  columns = ['Stock'], dtype = 'float64')
    ax = price_df.plot(colormap = 'cool', grid = False)
    ax.set_facecolor(color = 'k')
    ax.set_title('Impacted Stock Price')
    ax = plt.gca()
    yNumFmt = mticker.StrMethodFormatter('${x:,.2f}')
    ax.yaxis.set_major_formatter(yNumFmt)
    plt.plot(price_hist, 'o')
    plt.ylabel('Stock Price')
    plt.xlabel('Trade Number')
    plt.show()


def get_av_std(lq_time = 60, nm_trades = 60, tr_risk = 1e-6, trs = 100):
    
    # Create simulation environment
    env = sca.MarketEnvironment()

    # Reset the enviroment
    env.reset(liquid_time = lq_time, num_trades = nm_trades, lamb = tr_risk)

    # Get the trading list
    trl = env.get_trade_list()

    # Since we are not selling fractional shares we round up the shares in the trading list
    trade_list = round_trade_list(trl)

    # Set the initial shortfall to zero
    shortfall_hist = np.array([])

    for episode in range(trs):
        
        # Print current episode every 100 episodes
        if (episode + 1) % 100 == 0:
            print('Episode [{}/{}]'.format(episode + 1, trs), end = '\r', flush = True)
        
        # Reset the enviroment
        env.reset(seed = episode, liquid_time = lq_time, num_trades = nm_trades, lamb = tr_risk)

        # set the environment to make transactions
        env.start_transactions()

        for trade in trade_list:
            action = trade / env.shares_remaining
            _, _, _, info = env.step(action)

            if info.done:
                shortfall_hist = np.append(shortfall_hist, info.implementation_shortfall)
                break

    print('Average Implementation Shortfall: ${:,.2f}'.format(shortfall_hist.mean()))
    print('Standard Deviation of the Implementation Shortfall: ${:,.2f}'.format(shortfall_hist.std()))
    
    plt.plot(shortfall_hist, 'cyan', label='')
    plt.xlim(0, trs)
    ax = plt.gca()
    ax.set_facecolor('k')
    ax.set_xlabel('Episode', fontsize = 15)
    ax.set_ylabel('Implementation Shortfall (US $)', fontsize = 15)
    ax.axhline(shortfall_hist.mean(),0, 1, color = 'm', label='Average')
    yNumFmt = mticker.StrMethodFormatter('${x:,.0f}')
    ax.yaxis.set_major_formatter(yNumFmt)
    plt.legend()
    plt.show


import collections
import random

import numpy as np

ANNUAL_VOLAT = 0.12
BID_ASK_SP = 1 / 8
DAILY_TRADE_VOL = 5e6
TRAD_DAYS = 250
DAILY_VOLAT = ANNUAL_VOLAT / np.sqrt(TRAD_DAYS)
TOTAL_SHARES = 1000000
STARTING_PRICE = 50
LLAMBDA = 1e-6
LIQUIDATION_TIME = 60
NUM_N = 60
EPSILON = BID_ASK_SP / 2
SINGLE_STEP_VARIANCE = (DAILY_VOLAT * STARTING_PRICE) ** 2
ETA = BID_ASK_SP / (0.01 * DAILY_TRADE_VOL)
GAMMA = BID_ASK_SP / (0.1 * DAILY_TRADE_VOL)


class MarketEnvironment:
    def __init__(
        self, seed=0, lqd_time=LIQUIDATION_TIME, num_tr=NUM_N, llambda=LLAMBDA
    ):
        random.seed(seed)

        self.anv = ANNUAL_VOLAT
        self.basp = BID_ASK_SP
        self.dtv = DAILY_TRADE_VOL
        self.dpv = DAILY_VOLAT
        self.total_shares = TOTAL_SHARES
        self.starting_price = STARTING_PRICE
        self.llambda = llambda
        self.liquidation_time = lqd_time
        self.num_n = num_tr
        self.epsilon = EPSILON
        self.single_step_variance = SINGLE_STEP_VARIANCE
        self.eta = ETA
        self.gamma = GAMMA
        self.tau = self.liquidation_time / self.num_n
        self.eta_hat = self.eta - (0.5 * self.gamma * self.tau)

        self.kappa_hat = np.sqrt(
            (self.llambda * self.single_step_variance) / self.eta_hat
        )

        self.kappa = (
            np.arccosh((((self.kappa_hat**2) * (self.tau**2)) / 2) + 1) / self.tau
        )

        self.shares_remaining = self.total_shares
        self.time_horizon = self.num_n
        self.logReturns = collections.deque(np.zeros(6))
        self.prev_impact_price = self.starting_price
        self.transacting = False
        self.k = 0

    def reset(
        self, seed=0, liquid_time=LIQUIDATION_TIME, num_trades=NUM_N, lamb=LLAMBDA
    ):
        self.__init__(
            randomSeed=seed, lqd_time=liquid_time, num_tr=num_trades, lambd=lamb
        )

        self.initial_state = np.array(
            list(self.logReturns)
            + [
                self.time_horizon / self.num_n,
                self.shares_remaining / self.total_shares,
            ]
        )

        return self.initial_state

    def start_transactions(self):
        self.transacting = True
        self.tolerance = 1
        self.total_capture = 0
        self.prev_price = self.starting_price
        self.totalSSSQ = 0
        self.totalSRSQ = 0
        self.prev_utility = self.compute_ac_utility(self.total_shares)

    def step(self, action):
        class Info:
            pass

        info = Info()
        info.done = False

        if self.transacting and (
            self.time_horizon == 0 or abs(self.shares_remaining) < self.tolerance
        ):
            self.transacting = False

            info.done = True
            info.slippage = self.total_shares * self.starting_price - self.total_capture

            info.expected_shortfall = self.get_expected_shortfall(self.total_shares)

            info.expected_variance = (
                self.single_step_variance * self.tau * self.totalSRSQ
            )

            info.utility = (
                info.expected_shortfall + self.llambda * info.expected_variance
            )

        if self.k == 0:
            info.price = self.prev_impact_price
        else:
            info.price = self.prev_impact_price + np.sqrt(
                self.single_step_variance * self.tau
            ) * random.normalvariate(0, 1)

        if self.transacting:
            if isinstance(action, np.ndarray):
                action = action.item()

            leaves_now = self.shares_remaining * action

            if self.time_horizon < 2:
                leaves_now = self.shares_remaining

            info.share_to_sell_now = np.around(leaves_now)
            info.curr_perm_impact = self.perm_impact(info.share_to_sell_now)
            info.curr_temp_impact = self.temp_impact(info.share_to_sell_now)
            info.exec_price = info.price - info.curr_temp_impact
            self.total_capture += info.share_to_sell_now * info.exec_price

            self.logReturns.append(np.log(info.price / self.prev_price))
            self.logReturns.popleft()

            self.shares_remaining -= info.share_to_sell_now

            self.totalSSSQ += info.share_to_sell_now**2
            self.totalSRSQ += self.shares_remaining**2

            self.time_horizon -= 1
            self.prev_price = info.price
            self.prev_impact_price = info.price - info.curr_perm_impact

            curr_utility = self.compute_ac_utility(self.shares_remaining)
            reward = (abs(self.prev_utility) - abs(curr_utility)) / abs(
                self.prev_utility
            )
            self.prev_utility = curr_utility

            if self.shares_remaining <= 0:
                info.slippage = (
                    self.total_shares * self.starting_price - self.total_capture
                )

                info.done = True
        else:
            reward = 0.0

        self.k += 1

        state = np.array(
            list(self.logReturns)
            + [
                self.time_horizon / self.num_n,
                self.shares_remaining / self.total_shares,
            ]
        )

        return (state, np.array([reward]), info.done, info)

    def perm_impact(self, leaves):
        return self.gamma * leaves

    def temp_impact(self, leaves):
        return (self.epsilon * np.sign(leaves)) + ((self.eta / self.tau) * leaves)

    def get_expected_shortfall(self, leaves):
        return (
            0.5 * self.gamma * (leaves**2)
            + self.epsilon * leaves
            + (self.eta_hat / self.tau) * self.totalSSSQ
        )

    def get_ac_expected_shortfall(self, leaves):
        ft = 0.5 * self.gamma * (leaves**2)

        st = self.epsilon * leaves

        tt = self.eta_hat * (leaves**2)

        nft = np.tanh(0.5 * self.kappa * self.tau) * (
            self.tau * np.sinh(2 * self.kappa * self.liquidation_time)
            + 2 * self.liquidation_time * np.sinh(self.kappa * self.tau)
        )

        dft = 2 * (self.tau**2) * (np.sinh(self.kappa * self.liquidation_time) ** 2)

        fot = nft / dft

        return ft + st + (tt * fot)

    def get_ac_variance(self, leaves):
        ft = 0.5 * (self.single_step_variance) * (leaves**2)

        nst = self.tau * np.sinh(self.kappa * self.liquidation_time) * np.cosh(
            self.kappa * (self.liquidation_time - self.tau)
        ) - self.liquidation_time * np.sinh(self.kappa * self.tau)

        dst = (np.sinh(self.kappa * self.liquidation_time) ** 2) * np.sinh(
            self.kappa * self.tau
        )

        st = nst / dst

        return ft * st

    def compute_ac_utility(self, leaves):
        if self.liquidation_time == 0:
            return 0

        return self.get_ac_expected_shortfall(
            leaves
        ) + self.llambda * self.get_ac_variance(leaves)

    def get_trade_list(self):
        trade_list = np.zeros(self.num_n)
        ftn = 2 * np.sinh(0.5 * self.kappa * self.tau)
        ftd = np.sinh(self.kappa * self.liquidation_time)
        ft = (ftn / ftd) * self.total_shares

        for i in range(1, self.num_n + 1):
            st = np.cosh(self.kappa * (self.liquidation_time - (i - 0.5) * self.tau))
            trade_list[i - 1] = st

        trade_list *= ft

        return trade_list

    def observation_space_dimension(self):
        return 8

    def action_space_dimension(self):
        return 1

    def stop_transactions(self):
        self.transacting = False



import setuptools

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="acrl",
    version="0.0.3",
    author="Alexandre Brilhante",
    author_email="alexandre.brilhante@gmail.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/brilhana/almgren-chriss",
    project_urls={
        "Bug Tracker": "https://github.com/brilhana/almgren-chriss/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.8",
)
